@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/tests/progs/slapd-common.c,v 1.4.2.6 2008/02/11 23:26:50 kurt Exp $ */
+/* $OpenLDAP: pkg/ldap/tests/progs/slapd-common.c,v 1.4.2.8 2009/02/10 17:13:05 quanah Exp $ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1999-2008 The OpenLDAP Foundation.
+ * Copyright 1999-2009 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -40,8 +40,8 @@ tester_t progtype;
 
 #define	TESTER_SERVER_LAST	(LDAP_OTHER + 1)
 #define TESTER_CLIENT_LAST	(- LDAP_REFERRAL_LIMIT_EXCEEDED + 1)
-static unsigned ignore_server[ TESTER_SERVER_LAST ];
-static unsigned ignore_client[ TESTER_CLIENT_LAST ];
+static int ignore_server[ TESTER_SERVER_LAST ];
+static int ignore_client[ TESTER_CLIENT_LAST ];
 
 static struct {
 	char	*name;
@@ -126,8 +126,7 @@ static struct {
 static int
 tester_ignore_str2err( const char *err )
 {
-	int		i;
-	unsigned	ignore = 1;
+	int		i, ignore = 1;
 
 	if ( strcmp( err, "ALL" ) == 0 ) {
 		for ( i = 0; ignore_str2err[ i ].name != NULL; i++ ) {
@@ -147,6 +146,10 @@ tester_ignore_str2err( const char *err )
 	if ( err[ 0 ] == '!' ) {
 		ignore = 0;
 		err++;
+
+	} else if ( err[ 0 ] == '*' ) {
+		ignore = -1;
+		err++;
 	}
 
 	for ( i = 0; ignore_str2err[ i ].name != NULL; i++ ) {
@@ -183,24 +186,30 @@ tester_ignore_str2errlist( const char *err )
 	return 0;
 }
 
-unsigned
+int
 tester_ignore_err( int err )
 {
-	unsigned	rc = 1;
+	int rc = 1;
 
 	if ( err > 0 ) {
 		if ( err < TESTER_SERVER_LAST ) {
 			rc = ignore_server[ err ];
-			if ( rc ) {
+			if ( rc > 0 ) {
 				ignore_server[ err ]++;
+
+			} else if ( rc < 0 ) {
+				ignore_server[ err ]--;
 			}
 		}
 
 	} else if ( err < 0 ) {
 		if ( -err < TESTER_CLIENT_LAST ) {
 			rc = ignore_client[ -err ];
-			if ( rc ) {
+			if ( rc > 0 ) {
 				ignore_client[ -err ]++;
+
+			} else if ( rc < 0 ) {
+				ignore_server[ err ]--;
 			}
 		}
 	}
